import re
from typing import Iterable, Literal

import requests

from .logger import getLogger


LOGGER = getLogger('cloudns')


class ZoneNotFoundError(Exception):
    pass

class CouldNotGetIPError(Exception):
    pass

class APIError(Exception):
    pass

def get_current_ip(resolvers: Iterable[str]) -> str:
    '''
        Go through resolvers until one returns an IP.

        `resolvers` A list of IP resolvers.

        `log` A function which takes a string and logs it.
    '''

    for resolver in resolvers:
        response = requests.get(resolver, timeout = 5)

        if not response.ok:
            LOGGER.warning(f"[ERROR][{resolver}][{response.status_code}] {response.text}", extra = { 'resolver': resolver, 'status_code': response.status_code, 'content': response.content })

        current_ip = response.text.strip()

        # It suffices to check whether the search is not None since the regex matches from beginning to end.
        is_ipv4 = re.search(r'^((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)\.?\b){4}$', current_ip) is not None

        if is_ipv4:
            LOGGER.debug(f"[OK][{resolver}] Current IP: '{current_ip}'", extra = { 'resolver': resolver, 'current_ip': current_ip })
            return current_ip  # Break if we have found our IP

        LOGGER.warning(f"[WARNING][{resolver}] '{current_ip}' is not IPv4", extra = { 'resolver': resolver, 'ip': current_ip })

    raise CouldNotGetIPError('Could not get IP.')

def _cloudflare_api_request(method: Literal['GET', 'OPTIONS', 'HEAD', 'POST', 'PUT', 'PATCH', 'DELETE'], endpoint: str, api_key: str, headers: dict | None = None, timeout: int | None = 5, **kwargs) -> dict:
    response = requests.request(
        method,
        f'https://api.cloudflare.com/client/v4/{endpoint}',
        headers = {
            'Authorization': f'Bearer {api_key}',
            'Content-Type': 'application/json'
        } | (headers or {}),
        timeout = timeout,
        **kwargs
    )

    LOGGER.debug('Issued Cloudflare API request.', extra = { 'method': method, 'endpoint': endpoint, 'headers': headers, 'kwargs': kwargs, 'status_code': response.status_code, 'content': response.content })

    if not response.ok:
        LOGGER.error('', extra = { 'status_code': response.status_code, 'content': response.content })
        raise APIError(response.content)

    result = response.json()

    if not result['success']:
        LOGGER.error('API call was not successful.', extra = { 'status_code': response.status_code, 'content': response.content })
        raise APIError('API call was not successful.')

    return result['result']

def api_key_is_valid(api_key: str) -> bool:
    result = _cloudflare_api_request('GET', 'user/tokens/verify', api_key)
    return result['status'] == 'active'

def get_zone_id(api_key: str, domain: str) -> str:
    result = _cloudflare_api_request('GET', 'zones', api_key, params = { 'name': domain })

    if len(result) != 1:
        message = f"Expected a list of length '1' but got length '{len(result)}'"
        raise ZoneNotFoundError(message)

    return result[0]['id']

def get_A_records(api_key: str, zone_id: str) -> dict: # pylint: disable=invalid-name
    result = _cloudflare_api_request('GET', f'zones/{zone_id}/dns_records', api_key, params = { 'type': 'A' })
    return result

def create_A_record(api_key: str, zone_id: str, name: str, value: str) -> None: # pylint: disable=invalid-name
    _cloudflare_api_request(
        'POST',
        f'zones/{zone_id}/dns_records',
        api_key,
        json = {
            'type': 'A',
            'name': name,
            'content': value
        }
    )

def update_A_record(api_key: str, zone_id: str, record_id: str, value: str) -> None: # pylint: disable=invalid-name
    _cloudflare_api_request(
        'PATCH',
        f'/zones/{zone_id}/dns_records/{record_id}',
        api_key,
        json = { 'content': value }
    )
