import os
import sys
import argparse
import shutil
from typing import Any

from typing_extensions import Sequence

from .logger import setup_logging, getLogger
from .types import Config
from . import cloudns


def parse_args(args: Sequence[str]):

    class EnvDefault(argparse.Action):
        # Custom action to enable environment variables
        # https://stackoverflow.com/questions/10551117/setting-options-from-environment-variables-when-using-argparse/10551190#10551190
        def __init__(self, envvar: str, const: Any | None = None, default: Any = None, required: bool = True, **kwargs):
            if envvar in os.environ:
                default = os.environ[envvar] if const is None else const

            if const is not None:
                required = False
                kwargs['const'] = const
                kwargs['nargs'] = 0  # no additional arguments are expected

            # Make the argument optional if a default is determined
            if default is not None:
                required = False

            super().__init__(default=default, required=required, metavar=envvar, **kwargs)

        def __call__(self, parser, namespace, values, option_string=None):
            # If const is defined (store_const behavior), set it directly
            setattr(namespace, self.dest, self.const if self.const is not None else values)

    def formatter(prog):
        return argparse.ArgumentDefaultsHelpFormatter(prog, max_help_position=shutil.get_terminal_size().columns)

    parser = argparse.ArgumentParser(formatter_class=formatter)

    parser.add_argument(
        '-c', '--config', type = argparse.FileType('r'),
        action = EnvDefault, envvar = 'CLOUDNS_CONFIG',
        default = './config.json',
        help = 'Path to config file in JSON format.'
    )

    parser.add_argument(
        '-d', '--dry-run', type = bool,
        action = EnvDefault, envvar = 'CLOUDNS_DRY_RUN',
        const = True, default = False,
        help = 'Do not commit any changes'
    )

    return parser.parse_args(args)


def run(argv: Sequence[str]) -> None: # pylint: disable=too-many-locals
    # Parse command-line parameters
    args = parse_args(argv)  # pylint: disable=unused-variable

    config = Config.model_validate_json(args.config.read())

    setup_logging(config.log_path)
    logger = getLogger('cloudns')
    logger.debug('Started program.', extra = { 'argv': argv })

    try:
        current_ip = cloudns.get_current_ip(config.resolvers)

        for api_key, domains in config.api.items():
            if not cloudns.api_key_is_valid(api_key):
                logger.error(f"API key ending in '{api_key[-4:]}' is not active.")
                continue

            zones = {}
            for domain in domains:
                try:
                    zones[domain] = cloudns.get_zone_id(api_key, domain)
                except cloudns.ZoneNotFoundError as err:
                    logger.exception(err)

            records_to_update = {} # { ('sub_domain', 'domain'): ('zone_id', 'record_id' | None) }
            for domain, zone_id in zones.items():
                records = { record['name']: record for record in  cloudns.get_A_records(api_key, zone_id) }
                for sub_domain in domains[domain]:
                    full_domain = f'{sub_domain}.{domain}' if sub_domain != '@' else domain
                    if full_domain not in records:
                        logger.info(f"Could not find an A record for '{full_domain}'. Will be created.", extra = { 'sub': sub_domain, 'domain': domain })
                        records_to_update[(sub_domain, domain)] = (zone_id, None)
                        continue

                    if (record := records[full_domain])['content'] != current_ip:
                        logger.info(f"IP for '{full_domain}' does not match current IP. Will be updated.", extra = { 'sub': sub_domain, 'domain': domain, 'content': record['content'] })
                        records_to_update[(sub_domain, domain)] = (zone_id, record['id'])

            # Update/Create A Records #
            if args.dry_run:
                continue

            for (sub, domain), (zone_id, record_id) in records_to_update.items():
                full_domain = f'{sub}.{domain}' if sub != '@' else domain

                if record_id is None:
                    # Create record
                    cloudns.create_A_record(api_key, zone_id, full_domain, current_ip)
                    logger.info(f"A record for '{full_domain}' created.", extra = { 'sub': sub, 'domain': domain })
                    continue

                # Update Record
                cloudns.update_A_record(api_key, zone_id, record_id, current_ip)
                logger.info(f"'A record for '{full_domain}' updated.", extra = { 'sub': sub, 'domain': domain })

    except Exception as err:
        logger.exception(err)
        raise err from err


def main() -> None:
    run(sys.argv[1:])
